#include <cmath>
#include <functional>
#include <iostream>
#include <numbers>

#include <omp.h>

class MidsumException : std::exception {
    const char *what() const _GLIBCXX_TXN_SAFE_DYN _GLIBCXX_NOTHROW override {
        return "Invalid call to midsum function";
    }
};

auto midsum(std::function<double(double)> fun, double x1, double x2) -> double {
    if (x2 <= x1) {
        throw MidsumException();
    }
    auto dx = x2 - x1;
    return fun((x1 + (dx / 2))) * dx;
}

auto csin(double arg) -> double {
    return std::sin(arg) + 1;
}

auto main(int argc, char **argv) -> int {
    double midsum_val;
    double from = 0;
    double to = 2 * std::numbers::pi;
    int slices = 20;
    double step = 1.0 * (to - from) / slices;

    std::cout << "Num slices: " << slices << " Step size: " << step << std::endl;

#pragma omp parallel
    {
            #pragma omp for

            for (int i = 0; i < slices; i++) {
                std::cout << "Calculating step: " << i << " from " << i * step << " to " << (i + 1) * step << std::endl;
                midsum_val += midsum(csin, (i * step), ((i + 1) * step));
            }

    }

    std::cout << midsum_val <<
              std::endl;


    return 0;
}